/*
  exponentiation.c

  Uses the square and multiply algorithm for fast exponentation

  This works but is quite inefficient:
  1. This could take advantage of fermat's little theorem to speed the
     exponentiation, expecially for the case of RSA decryption.
  2. The squares in the algorithm will begin to repeat. Rather than
     simply calculating them over and over, some form of array/vector/list
     should be used to lookup the repeating values.

  Part of my implementation of the RSA scheme
*/

/*
  Evaluate p^q (mod m) quickly

  Algorithm:
  1. Write q in binary form: q = (an)2^n + (a(n-1))2^(n-1) + ... + (a2)2^2 + (a1)2^1
  2. Compute p^1, p^2, p^4, ..., p^(2^n) (mod m) by
     squaring the previous term in the sequence
  3. Multiply together all a^(2^i) such that (ai) = 1 to get the result
*/

#include "exponentiation.h"

unsigned long int exponentiate (unsigned long int p, unsigned long int q,
				unsigned long int m)
{
  unsigned long int square = 0, result = 0;

  /* m must be less that 65536 or else we may get an overflow situation */
  //assert (m < 65536);
  
  /* Do the first step */
  square = p % m;
  if (q & 0x1)
    {
      result = square;
    }
  q = q >> 1;
  
  /* Continue until there are no more 1's left in q */
  while (q != 0)
    {
      /* Square the previous term and evaluate the mod */
      square = (square * square) % m;

      /* Update the result */
      if (q & 0x1)
	{
	  if (result == 0)
	    {
	      result = square;
	    }
	  else
	    {
	      result = (result * square) % m;
	    }
	}

      /* Bitshift the exponent */
      q = q >> 1;
    }

  return result;
}
	  
	      
	  
      
      
      
  
